"use strict";

const {
  ActionTransport
} = require('@microfleet/core');

const Promise = require('bluebird');

const Errors = require('common-errors');

const moment = require('moment');

const noop = require('lodash/noop');

const is = require('is');

const scrypt = require('../utils/scrypt');

const redisKey = require('../utils/key');

const jwt = require('../utils/jwt');

const isActive = require('../utils/is-active');

const isBanned = require('../utils/is-banned');

const handlePipeline = require('../utils/pipeline-error');

const {
  checkMFA
} = require('../utils/mfa');

const {
  verifySignedToken
} = require('../auth/oauth/utils/get-signed-token');

const {
  USERS_ACTION_DISPOSABLE_PASSWORD,
  USERS_DISPOSABLE_PASSWORD_MIA,
  USERS_ID_FIELD,
  USERS_USERNAME_FIELD,
  USERS_MFA_FLAG,
  MFA_TYPE_OPTIONAL,
  USERS_INVALID_TOKEN,
  ErrorUserNotFound
} = require('../constants');
/**
 * Internal functions
 */


const is404 = e => parseInt(e.message, 10) === 404;

const globalLoginAttempts = async ctx => {
  const {
    config
  } = ctx;
  const pipeline = ctx.redis.pipeline();
  const globalRemoteIpKey = ctx.globalRemoteIpKey = redisKey('gl!ip!ctr', ctx.remoteip);
  pipeline.incrby(globalRemoteIpKey, 1);

  if (config.keepGlobalLoginAttempts > 0) {
    pipeline.expire(globalRemoteIpKey, config.keepGlobalLoginAttempts);
  }

  const [globalAttempts] = await pipeline.exec().then(handlePipeline);
  ctx.globalLoginAttempts = globalAttempts; // globally scoped go next

  if (globalAttempts > ctx.globalLockAfterAttempts) {
    const duration = moment().add(config.keepGlobalLoginAttempts, 'seconds').toNow(true);
    const msg = `You are locked from making login attempts for the next ${duration} from ipaddress '${ctx.remoteip}'`;
    throw new Errors.HttpStatusError(429, msg);
  }
};

const localLoginAttempts = async (ctx, data) => {
  const {
    config
  } = ctx;
  const pipeline = ctx.redis.pipeline();
  const userId = data[USERS_ID_FIELD];
  const remoteipKey = ctx.remoteipKey = redisKey(userId, 'ip', ctx.remoteip);
  pipeline.incrby(remoteipKey, 1);

  if (config.keepLoginAttempts > 0) {
    pipeline.expire(remoteipKey, config.keepLoginAttempts);
  }

  const [scopeAttempts] = await pipeline.exec().then(handlePipeline);
  ctx.loginAttempts = scopeAttempts; // locally scoped login attempts are prioritized

  if (scopeAttempts > ctx.lockAfterAttempts) {
    const duration = moment().add(config.keepLoginAttempts, 'seconds').toNow(true);
    const msg = `You are locked from making login attempts for the next ${duration} from ipaddress '${ctx.remoteip}'`;
    throw new Errors.HttpStatusError(429, msg);
  }
};
/**
 * Checks if login attempts from remote ip have exceeded allowed
 * limits
 */


async function checkLoginAttempts(dataOrError) {
  const promises = [globalLoginAttempts(this)];

  if (dataOrError instanceof Error === false) {
    promises.push(localLoginAttempts(this, dataOrError));
  }

  await Promise.all(promises);
}
/**
 * Verifies passed hash
 */


function verifyHash({
  password
}, comparableInput) {
  return scrypt.verify(password, comparableInput);
}

async function verifyOAuthToken({
  id
}, token) {
  const providerData = await verifySignedToken.call(this, token);

  if (providerData.profile.userId !== id) {
    throw USERS_INVALID_TOKEN;
  }

  return true;
}
/**
 * Checks onу-time password
 */


async function verifyDisposablePassword(ctx, data) {
  try {
    return await ctx.tokenManager.verify({
      action: USERS_ACTION_DISPOSABLE_PASSWORD,
      id: data[USERS_USERNAME_FIELD],
      token: ctx.password
    });
  } catch (e) {
    if (is404(e)) {
      throw USERS_DISPOSABLE_PASSWORD_MIA;
    }

    throw e;
  }
}
/**
 * Determines which strategy to use
 */


function getVerifyStrategy(data) {
  if (this.isSSO === true) {
    return null;
  }

  if (is.string(this.password) !== true || this.password.length < 1) {
    throw new Errors.HttpStatusError(400, 'should supply password');
  }

  if (this.isDisposablePassword === true) {
    return verifyDisposablePassword(this, data);
  }

  if (this.isOAuthFollowUp === true) {
    return verifyOAuthToken.call(this.service, data, this.password);
  }

  return verifyHash(data, this.password);
}
/**
 * Drops login attempts counter
 */


function dropLoginCounter() {
  this.loginAttempts = 0;
  this.globalLoginAttempts = 0;
  return this.redis.dropLoginCounter(2, this.remoteipKey, this.globalRemoteIpKey);
}
/**
 * Returns user info
 */


async function getUserInfo(internalData) {
  const datum = await Promise.bind(this.service, [internalData.id, this.audience]).spread(jwt.login); // NOTE: transformed to boolean

  datum.mfa = !!internalData[USERS_MFA_FLAG];
  return datum;
}
/**
 * Enriches error with amount of login attempts
 */


function enrichError(err) {
  if (this.remoteip) {
    err.loginAttempts = this.loginAttempts;
    err.globalLoginAttempts = this.globalLoginAttempts;
  }

  throw err;
}

function verifyInternalData(data) {
  if (!data) throw ErrorUserNotFound;
}
/**
 * @api {amqp} <prefix>.login User Authentication
 * @apiVersion 1.0.0
 * @apiName LoginUser
 * @apiGroup Users
 *
 * @apiDescription Provides various strategies for user authentication. Returns signed JWT token that could be used
 * for state resolution and authorization, as well as user object
 *
 * @apiParam (Payload) {String} username - currently only email
 * @apiParam (Payload) {String} password - plain text password, will be compared to store hash
 * @apiParam (Payload) {String} audience - metadata to be returned, as well embedded into JWT token
 * @apiParam (Payload) {String} [remoteip] - security logging feature, not used
 * @apiParam (Payload) {String} [isDisposablePassword=false] - use disposable password for verification
 * @apiParam (Payload) {String} [isSSO=false] - verification was already performed by single sign on (ie, facebook)
 */


function login({
  params,
  locals
}) {
  const config = this.config.jwt;
  const {
    redis,
    tokenManager
  } = this;
  const {
    isOAuthFollowUp,
    isDisposablePassword,
    isSSO,
    password
  } = params;
  const {
    lockAfterAttempts,
    globalLockAfterAttempts,
    defaultAudience
  } = config;
  const audience = params.audience || defaultAudience;
  const remoteip = params.remoteip || false;
  const verifyIp = remoteip && lockAfterAttempts > 0; // build context

  const ctx = {
    // service data
    service: this,
    tokenManager,
    redis,
    config,
    // counters
    loginAttempts: 0,
    globalLoginAttempts: 0,
    // business logic params
    params,
    isOAuthFollowUp,
    isDisposablePassword,
    isSSO,
    password,
    lockAfterAttempts,
    globalLockAfterAttempts,
    audience,
    remoteip,
    verifyIp
  };
  return Promise.bind(ctx, locals.internalData) // verify that locals.internalData exists
  .tap(verifyInternalData) // record global login attempt even on 404
  .tapCatch(verifyIp ? checkLoginAttempts : noop).tap(verifyIp ? checkLoginAttempts : noop) // different auth strategies
  .tap(getVerifyStrategy) // pass-through or drop counter
  .tap(verifyIp ? dropLoginCounter : noop) // do verifications on the logged in account
  .tap(isActive).tap(isBanned) // fetch final user information
  .then(getUserInfo) // enriches and rethrows
  .catch(enrichError);
}

login.mfa = MFA_TYPE_OPTIONAL;
login.allowed = checkMFA;
login.transports = [ActionTransport.amqp, ActionTransport.internal];
module.exports = login;