"use strict";

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

const Errors = require('common-errors');

const Promise = require('bluebird');

const jwt = Promise.promisifyAll(require('jsonwebtoken')); // internal modules

const redisKey = require('./key');

const getMetadata = require('../utils/get-metadata');

const {
  verify: verifyHMAC
} = require('./signatures');

const {
  USERS_TOKENS,
  USERS_API_TOKENS,
  USERS_ID_FIELD,
  USERS_INVALID_TOKEN,
  USERS_USERNAME_FIELD,
  USERS_AUDIENCE_MISMATCH,
  USERS_MALFORMED_TOKEN,
  BEARER_USERNAME_FIELD,
  BEARER_LEGACY_USERNAME_FIELD
} = require('../constants'); // cache this to not recreate all the time


const mapJWT = props => ({
  jwt: props.jwt,
  user: {
    [USERS_ID_FIELD]: props.userId,
    metadata: props.metadata
  }
});

function verifyData(token, tokenOptions, extraOpts = {}) {
  return jwt.verifyAsync(token, tokenOptions.secret, _objectSpread({}, tokenOptions.extra, {}, extraOpts, {
    issuer: tokenOptions.issuer,
    algorithms: [tokenOptions.hashingFunction]
  }));
}
/**
 * Logs user in and returns JWT and User Object
 * @param  {String}  username
 * @param  {String}  _audience
 * @return {Promise}
 */


exports.login = function login(userId, _audience) {
  const {
    redis,
    config,
    flake
  } = this;
  const {
    jwt: jwtConfig
  } = config;
  const {
    hashingFunction: algorithm,
    defaultAudience,
    secret
  } = jwtConfig;
  let audience = _audience || defaultAudience; // will have iat field, which is when this token was issued
  // we can check last access and verify the expiration date based on it

  const payload = {
    [USERS_USERNAME_FIELD]: userId,
    cs: flake.next()
  };
  const token = jwt.sign(payload, secret, {
    algorithm,
    audience,
    issuer: 'ms-users'
  });

  if (audience !== defaultAudience) {
    audience = [audience, defaultAudience];
  } else {
    audience = [audience];
  }

  return Promise.props({
    lastAccessUpdated: redis.zadd(redisKey(userId, USERS_TOKENS), Date.now(), token),
    jwt: token,
    userId,
    metadata: getMetadata.call(this, userId, audience)
  }).then(mapJWT);
};
/**
 * Logs error & then throws
 */


function remapInvalidTokenError(err) {
  this.log.debug('error decoding token', err);
  return Promise.reject(USERS_INVALID_TOKEN);
}
/**
 * Erases the token
 */


function eraseToken(decoded) {
  return this.redis.zrem(redisKey(decoded[USERS_USERNAME_FIELD], USERS_TOKENS), this.token);
}
/**
 * Removes token if it is valid
 * @param  {String} token
 * @param  {String} audience
 * @return {Promise}
 */


exports.logout = function logout(token, audience) {
  const {
    redis,
    config
  } = this;
  return verifyData(token, config.jwt, {
    audience
  }).bind(this).catch(remapInvalidTokenError).bind({
    redis,
    token
  }).then(eraseToken).return({
    success: true
  });
};
/**
 * Removes all issued tokens for a given user
 * @param {String} username
 */


exports.reset = function reset(userId) {
  return this.redis.del(redisKey(userId, USERS_TOKENS));
};
/**
 * Parse last access
 */


function getLastAccess(_score) {
  // parseResponse
  const score = _score * 1; // throw if token not found or expired

  if (_score == null || Date.now() > score + this.ttl) {
    throw new Errors.HttpStatusError(403, 'token has expired or was forged');
  }

  return score;
}
/**
 * Refreshes last access token
 */


function refreshLastAccess() {
  return this.redis.zadd(this.tokensHolder, Date.now(), this.token);
}
/**
 * Verify decoded token
 */


async function verifyDecodedToken(decoded) {
  if (this.audience.indexOf(decoded.aud) === -1) {
    return Promise.reject(USERS_AUDIENCE_MISMATCH);
  }

  const username = decoded[USERS_USERNAME_FIELD];
  const tokensHolder = this.tokensHolder = redisKey(username, USERS_TOKENS);
  const score = await this.redis.zscore(tokensHolder, this.token);
  const lastAccess = await getLastAccess.call(this, score);

  if (!this.peek) {
    await refreshLastAccess.call(this, lastAccess);
  }

  return decoded;
}
/**
 * Verifies token and returns decoded version of it
 * @param  {String} token
 * @param  {Array} audience
 * @param  {Boolean} peek
 * @param  {String} [overrideSecret]
 * @return {Promise}
 */


exports.verify = function verifyToken(token, audience, peek) {
  const jwtConfig = this.config.jwt;
  const ctx = {
    audience,
    token,
    peek,
    redis: this.redis,
    log: this.log,
    ttl: jwtConfig.ttl
  };
  return verifyData(token, jwtConfig).bind(ctx).catch(remapInvalidTokenError).then(verifyDecodedToken);
};
/**
 * Verifies internal token
 * @param {String} token
 * @param {Array} audience
 * @param {Boolean} peek
 * @return {Promise}
 */


exports.internal = async function verifyInternalToken(token) {
  const tokenParts = token.split('.');

  if (tokenParts.length !== 3) {
    return Promise.reject(USERS_MALFORMED_TOKEN);
  }

  const [userId, uuid, signature] = tokenParts; // token is malformed, must be username.uuid.signature

  if (!userId || !uuid || !signature) {
    return Promise.reject(USERS_MALFORMED_TOKEN);
  } // md5 hash


  const isLegacyToken = userId.length === 32 && /^[a-fA-F0-9]{32}$/.test(userId); // this is needed to pass ctx of the

  const payload = `${userId}.${uuid}`;
  const isValid = verifyHMAC.call(this, payload, signature);

  if (!isValid) {
    return Promise.reject(USERS_MALFORMED_TOKEN);
  } // at this point signature is valid and we need to verify that it was not
  // erase or expired


  const key = redisKey(USERS_API_TOKENS, payload);
  const tokenField = isLegacyToken ? BEARER_LEGACY_USERNAME_FIELD : BEARER_USERNAME_FIELD;
  const id = await this.redis.hget(key, tokenField);
  return {
    [tokenField]: id
  };
};
/**
 * Sign data
 * @param  {any} paylaod
 * @param  {Object} tokenOptions
 * @return {Promise}
 */


exports.signData = function signData(payload, tokenOptions) {
  const {
    hashingFunction: algorithm,
    secret,
    issuer,
    extra
  } = tokenOptions;
  return jwt.sign(payload, secret, _objectSpread({}, extra, {
    algorithm,
    issuer
  }));
};
/**
 * Verify data
 * @param  {String} token
 * @param  {Object} tokenOptions
 * @return {Promise}
 */


exports.verifyData = verifyData;